#include <windows.h>
#include <cstdio>

using namespace std;

void TrazarTexto(int x, int y, const char *texto) {
    HANDLE hMutex;
    COORD dwPos = {x, y};

    hMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, "MutexAccesoPantalla");

    if(WaitForSingleObject(hMutex, INFINITE) == WAIT_OBJECT_0){
        SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), dwPos);
        printf("%s", texto);
        ReleaseMutex(hMutex);
    }
    CloseHandle(hMutex);
}

class Hora {
    public:
        Hora(int xi, int yi) : x(xi), y(yi), seg(0) {
            manipulador = CreateThread( 0, 0, hiloMostrarHora, (void*)this, 0, 0);
            GetLocalTime(&st);
        }
        ~Hora() {
            WaitForSingleObject(manipulador, INFINITE);
            CloseHandle(manipulador);
        }
        void Terminar() { salir = true; }
    private:
        void Mostrar() const;
        void EsperarCambio();
        void ActualizarSeg();
        bool Salir() const { return salir; }
        SYSTEMTIME st;                    // Estructura para obtener la hora
        int x;
        int y;
        int seg;
        HANDLE manipulador;
        static bool salir;
        static DWORD WINAPI hiloMostrarHora( void* );
};

class TextoRotativo {
    public:
        TextoRotativo(int xc, int yc, int anchoc, int retc, const char *cad);
        ~TextoRotativo();
        void Terminar() { salir = true; }

    private:
        void Mostrar();
        void IncrementaPos();
        void Rotar();
        void Esperar() { Sleep(retardo); }
        bool Salir() const { return salir; }
        bool Reservar();
        void Liberar();
        char *cadena;
        int longitud;
        int pos;
        int x;
        int y;
        int ancho;
        int retardo;
        HANDLE manipulador;
        HANDLE hMutex;
        static bool salir;
        static DWORD WINAPI hiloRotarTexto( void* par );
};

// Definir clases:

void Hora::Mostrar() const {
    char texto[9]; // Variable con el texto de la hora

    sprintf(texto, "%02d:%02d:%02d", st.wHour, st.wMinute, st.wSecond);
    TrazarTexto(x, y, texto);
}

void Hora::EsperarCambio() {
    while(seg == st.wSecond) {
        GetLocalTime(&st);
        Sleep(100);
    }
}

void Hora::ActualizarSeg() {
    seg = st.wSecond;
}

DWORD WINAPI Hora::hiloMostrarHora( void* par) {
    Hora *hora = (Hora*)par;  // Recuperar el parmetro, un puntero a Hora

    do {
        hora->EsperarCambio();
        hora->Mostrar();
        hora->ActualizarSeg();
    } while(!hora->Salir()); // Repetir hasta que salir sea true

    return 0;
}

TextoRotativo::TextoRotativo(int xc, int yc, int anchoc, int retc, const char *cad) :
    pos(0), x(xc), y(yc), ancho(anchoc), retardo(retc) {
    hMutex = OpenMutex(MUTEX_ALL_ACCESS, FALSE, "MutexAccesoPantalla");
    manipulador = CreateThread( 0, 0, hiloRotarTexto, (void*)this, 0, 0);
    longitud = strlen(cad);
//            if(longitud < ancho) longitud = ancho;
    cadena = new char[longitud+1];
    strcpy(cadena, cad);
}

TextoRotativo::~TextoRotativo() {
    WaitForSingleObject(manipulador, INFINITE);
    CloseHandle(manipulador);
    CloseHandle(hMutex);
    delete[] cadena;
}

void TextoRotativo::Rotar() {
    SMALL_RECT re1 = {x+1, y, x+ancho, y};
    COORD coor = {x, y};
    CHAR_INFO ci = { {(pos+ancho) >= longitud ? cadena[pos+ancho-longitud] : cadena[pos+ancho]},
        FOREGROUND_BLUE+FOREGROUND_GREEN+FOREGROUND_RED};

    ScrollConsoleScreenBuffer(GetStdHandle(STD_OUTPUT_HANDLE), &re1, 0, coor, &ci);
}

bool TextoRotativo::Reservar() {
    if(WaitForSingleObject(hMutex, INFINITE) == WAIT_OBJECT_0) return true;
    return false;
}

void TextoRotativo::Liberar() {
    ReleaseMutex(hMutex);
}

void TextoRotativo::Mostrar() {
    COORD dwPos = {x, y};
    int n = 0;
    int m;

    SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), dwPos);
    n = ancho < longitud-pos ? ancho : longitud-pos;
    WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), cadena+pos, n, 0, 0);
    if(ancho-n > 0)
       WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE), cadena, ancho-n, 0, 0);
}

void TextoRotativo::IncrementaPos() {
    pos++;
    if(!cadena[pos]) pos = 0;
}

DWORD WINAPI TextoRotativo::hiloRotarTexto( void* par) {
    TextoRotativo *texto = (TextoRotativo*)par;

    // Texto inicial
    if(texto->Reservar()) {
        texto->Mostrar();
        texto->Liberar();
    }
    do {
        // Pintar 'ancho' caracteres a partir de (x,y), empezando en la
        // posicin 'pos' de la 'cadena':
        if(texto->Reservar()) {
            //texto->Rotar();
            texto->Mostrar();
            texto->Liberar();
            texto->IncrementaPos();
            texto->Esperar();
        }
    } while(!texto->Salir());
    return 0;
}

bool Hora::salir = false;
bool TextoRotativo::salir = false;

int main() {
    HANDLE hMutex = CreateMutex(NULL, FALSE, "MutexAccesoPantalla");
    SetConsoleOutputCP(1252);

    Hora hora(70, 1);
    Hora hora2(1, 1);
    Hora hora3(70, 23);
    Hora hora4(1, 23);
    TextoRotativo texto(20,2,15,100, "Texto rotativo de prueba ------ ");
    TextoRotativo texto2(20, 15, 10, 50, "Articulos Con Clase - Salvador Pozo - ");
    TextoRotativo texto3(5, 20, 70, 60, "Generalmente tendemos a pensar que la multitarea se limita a poder tener abiertas varias ventanas "
        "con distintas aplicaciones a la vez, pero hay mucho ms. Por ejemplo, un navegador web, puede "
        "tener abiertas varias pginas a la vez. A su vez, cada pgina, mientras se carga, puede estar descargando "
        "varios recursos: texto, imgenes, etc, y actualizando la pantalla a medida que los descarga. ------------------ ");
    TextoRotativo texto4(30, 0, 20, 150, "Pulsa una tecla para salir del programa.    ----   ");

    // Esperar a que el usuario pulse una tecla, el hilo se sigue ejecutando:
    getchar();
    hora.Terminar();
    texto.Terminar();

	// Liberar manipuladores:
    CloseHandle(hMutex);
    return 0;
}
